/*
 * Copyright (C) 2001-2003 by egnite Software GmbH. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the copyright holders nor the names of
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY EGNITE SOFTWARE GMBH AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL EGNITE
 * SOFTWARE GMBH OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF
 * THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * For additional information see http://www.ethernut.de/
 *
 */

/*
 * $Log: avr_nutinit.c,v $
 * Revision 1.10  2004/09/01 14:27:03  haraldkipp
 * Using configuration values from cfg/memory.h.
 * Added configurable reserved memory area.
 * Automatic check for external memory removed.
 *
 * Revision 1.9  2004/07/28 13:43:25  drsung
 * Corrected a misplaced #endif after last change.
 *
 * Revision 1.8  2004/07/09 19:51:17  freckle
 * Added new function NutThreadSetSleepMode to tell nut/os to set the MCU
 * into sleep mode when idle (avr-gcc && avr128 only)
 *
 * Revision 1.7  2004/07/09 14:40:43  freckle
 * Moved ((volatile u_char *) NUTRAMEND) cast into NUTRAMENDPTR define
 *
 * Revision 1.6  2004/07/09 14:23:13  freckle
 * Allow setting of NUTRAMEND by giving it as a compiler flag
 *
 * Revision 1.5  2004/05/25 17:13:48  drsung
 * Bit name SRW10 is not defined for atmega103, so added some defines to make it compatible.. :-X
 *
 * Revision 1.4  2004/05/25 12:03:37  olereinhardt
 * Sorry, fixed typing bug
 *
 * Revision 1.3  2004/05/25 12:00:37  olereinhardt
 * Newly added 3Waitstate support now needs to be enabled by
 * defining NUT_3WAITSTATES. By default this behaves like normal
 *
 *
 * Revision 1.1  2004/03/16 16:48:46  haraldkipp
 * Added Jan Dubiec's H8/300 port.
 *
 *
 */

#include <cfg/memory.h>

#ifdef NUTXMEM_SIZE
/*!
 * \brief Last memory address using external SRAM.
 */
#define NUTMEM_END (u_short)(NUTXMEM_START + (u_short)NUTXMEM_SIZE - 1U)

#else
/*!
 * \brief Last memory address without using external SRAM.
 *
 * \todo Shall we support NUTRAMEND for backward compatibility? If, then
 *       let's do it in cfg/memory.h.
 */
#define NUTMEM_END (u_short)(NUTMEM_START + (u_short)NUTMEM_SIZE - 1U)

#endif

#ifdef NUTMEM_RESERVED
/*!
 * \brief Number of bytes reserved in on-chip memory.
 *
 * AVR offers the option to temporarly use address and data bus
 * lines as general purpose I/O. If such drivers need data memory,
 * this must be located at internal memory addresses.
 * 
 * \todo Not a nice implementation but works as long as this module
 *       is linked first. Should be made a linker option.
 */
u_char nutmem_onchip[NUTMEM_RESERVED];
#endif

/* sleep mode to put avr in idle thread, SLEEP_MODE_NONE is used for for non sleeping */
#ifdef __GNUC__
u_char idle_sleep_mode = SLEEP_MODE_NONE;
#endif

#ifdef __GNUC__
/*
 * Some special declarations for AVRGCC. The runtime library
 * executes section .init8 before finally jumping to main().
 * We never let it do that jump, but start main() as a
 * separate thread. This introduces new problems:
 * 1. The compiler reinitializes the stack pointer when
 *    entering main, at least version 3.3 does it.
 * 2. The compiler doesn't allow to redeclare main to make
 *    it fit for NutThreadCreate().
 * 3. The runtime library requires, that main is defined
 *    somewhere.
 * Solutions:
 * 1. We do not use main at all, but let the preprocessor
 *    redefine it to NutAppMain() in the application code.
 *    See compiler.h. Note, that the declaration of NutAppMain 
 *    in this module differs from the one in the application
 *    code. Fortunately the linker doesn't detect this hack.
 * 2. We use a linker option to set the symbol main to zero.
 *
 * Thanks to Joerg Wunsch, who helped to solve this.
 */
void NutInit(void) __attribute__ ((naked)) __attribute__ ((section(".init8")));
extern void NutAppMain(void *arg) __attribute__ ((noreturn));
#else
extern void main(void *);
#endif

#if defined(__GNUC__) && defined(NUTXMEM_SIZE)

static void NutInitXRAM(void) __attribute__ ((naked, section(".init1"), used));
void NutInitXRAM(void)
{
    /* At the very beginning enable extended memory interface.
     */
#ifdef __AVR_ATmega128__    
    MCUCR = _BV(SRE) | _BV(SRW10);
#ifdef NUT_3WAITSTATES
    outb (XMCRA, 0x42);
    outb (XMCRB, 0x00);
#endif
#else
    MCUCR = _BV(SRE) | _BV(SRW);
#endif
}

#endif

/*!
 * \addtogroup xgThread
 */
/*@{*/

/*! \fn NutThreadSetSleepMode(u_char mode)
 * \brief Sets the sleep mode to enter in Idle thread
 *
 * If the idle thread is running, no other thread is active
 * so we can safely put the mcu to sleep.
 * 
 * \param mode one of the sleep modes defined in avr/sleep.h or
 *             sleep_mode_none (don't enter sleep mode)  
 */
#ifdef __GNUC__
void NutThreadSetSleepMode(u_char mode)
{
    idle_sleep_mode = mode;
}
#endif

/*!
 * \addtogroup xgNutInit
 */
/*@{*/

/*! \fn NutIdle(void *arg)
 * \brief Idle thread. 
 *
 * After initializing the timers, the idle thread switches to priority 254
 * and enters an endless loop.
 *
 * \todo Make the stack size of the application main thread configurable.
 */
THREAD(NutIdle, arg)
{
#ifdef __GNUC__
    u_char sleep_mode;
#endif    
    /* Initialize system timers. */
    NutTimerInit();

    /* Create the main application thread. */
    NutThreadCreate("main", main, 0, 768);

    /*
     * Run in an idle loop at the lowest priority. We can still
     * do something useful here, like killing terminated threads
     * or putting the CPU into sleep mode.
     */
    NutThreadSetPriority(254);
    for (;;) {
        NutThreadYield();
        NutThreadDestroy();
        
#ifdef __GNUC__
        if (idle_sleep_mode != SLEEP_MODE_NONE) {
            sleep_mode = MCUCR & SLEEP_MODE_EXT_STANDBY;
            set_sleep_mode(idle_sleep_mode);
            sleep_mode();
            set_sleep_mode(sleep_mode);
        }
#endif
    }
}

/*!
 * \brief Nut/OS Initialization.
 *
 * Initializes the memory management and the thread system and starts 
 * an idle thread, which in turn initializes the timer management. 
 * Finally the application's main() function is called.
 *
 * Depending on the compiler, different methods are used to execute this
 * function before main() is called.
 *
 * For ICCAVR the default crtatmega.o startup file is replaced by
 * crtnut.o, which calls NutInit instead of main(). This is done
 * by adding the following compiler options in the project:
 * \code -ucrtnut.o nutinit.o \endcode
 * 
 * For AVRGCC this function is located in section .init8, which is
 * called immediately before jumping to main(). NutInit is defined
 * as:
 * \code
 * void NutInit(void) __attribute__ ((naked)) __attribute__ ((section (".init8")));
 * \endcode
 *
 * \todo Verify ICCAVR external memory handling.
 *
 * \todo Make heap threshold configurable, currently hardcoded at 384.
 *
 * \todo Make stack size of the idle thread configurable.
 *
 * \todo Make wait states for external memory access configuratble.
 *
 * \todo Make early UART initialization for kernel debugging configurable.
 */
void NutInit(void)
{
    /*
     * We can't use local variables in naked functions.
     */
#ifdef NUTDEBUG
    outp(7, UBRR);
    outp(BV(RXEN) | BV(TXEN), UCR);
#endif

#ifndef __GNUC__
// FIXME: move this line to an appropriate initialization section of ICCAVR (os)
#ifdef NUTXMEM_SIZE
    outp(BV(SRE) | BV(SRW), MCUCR);
#endif
#endif

    /* If we have external RAM, initialize stack pointer to
     *  end of external RAM to avoid overwriting .data and .bss section
     */
    SP = (u_short)(NUTMEM_END);

    /* Then add the remaining RAM to heap.
     *
     * 20.Aug.2004 haraldkipp: This had been messed up somehow. It's nice to have 
     * one continuous heap area, but we lost the ability to have systems with
     * a gap between internal and external RAM.
     */
    if ((u_short)NUTMEM_END - (u_short) (&__heap_start) > 384) {
        NutHeapAdd(&__heap_start, (u_short) NUTMEM_END - 256 - (u_short) (&__heap_start));
    }

    /*
     * Read eeprom configuration.
     */
    if (NutLoadConfig()) {
        strcpy(confos.hostname, "ethernut");
        NutSaveConfig();
    }

    /* Create idle thread
     */
    NutThreadCreate("idle", NutIdle, 0, 384);
}

/*@}*/
